#!/usr/bin/env python2
from __future__ import print_function

__version__   = "2019-05-22"
__author__    = "ed <irc.rizon.net>"
__url__       = "https://ocv.me/dev/?sparsemulti.py"
__credits__   = ["stackoverflow.com"]
__license__   = "wtfpl"


"""
creates a sparse file and uses multiple writers
(threads or subprocesses) to write segments in parallel

tested on:
  ext2, linux 4.19.41 (alpine), py2.7.16, py3.6.8
  apfs, osx 18.6.0 (mojave), py2.7.10, py3.7.2
  ext4, linux 3.10.73 (android 8.1 @ nexus 5x), py3.7.3
  ntfs, windows xp sp3, py2.7.16

on winxp,
  f.truncate(total_sz) spawns file instantly, however
  f.close() takes 4sec on 2gb after any write near the end,
  so appears to non-sparse all leading zeros before 1st write,
  otherwise pass  (sidenote: vbox vdi does not grow)
"""


import time
import struct
import argparse


def gendata(v0, nbytes):
	"""
	yields 2048-byte chunks of uints counting from v0,
	mod-4 nbytes will be produced in total
	"""
	nwords = nbytes // 4
	csz = 512
	while csz > 0:
		while nwords >= csz:
			yield struct.pack('>'+'L'*csz, *range(v0, v0+csz))
			nwords -= csz
			v0 += csz
		csz = nwords


def writer(sz, v0, ofs):
	"""
	opens gen.out for read/write (must exist),
	seeks to ofs and writes sz uints starting from v0
	"""
	with open('gen.out', 'rb+') as f:
		f.seek(ofs)
		print('hi')
		for blk in gendata(v0, sz):
			f.write(blk)


if __name__ == '__main__':
	ap = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
	ap.add_argument('-s', metavar='MiB', help='num megabytes to write per worker', type=int, default=256)
	ap.add_argument('worker_type', choices=['thr','mp'], help='use threads or multiprocessing', type=str)
	ar = ap.parse_args()
	
	if ar.worker_type == 'thr':
		import threading
		
		def make_worker(**args):
			rv = threading.Thread(**args)
			rv.daemon = True
			return rv
		
		def is_busy():
			for worker in workers:
				if worker.isAlive:
					return True
			
			return False
	
	else:
		import multiprocessing as mp
		
		def make_worker(**args):
			return mp.Process(**args)
		
		def is_busy():
			# https://stackoverflow.com/questions/19353677/multiprocessing-process-is-alive-returns-true-although-process-has-finished-w
			for worker in workers:
				if worker.is_alive():
					return True
			
			return False
	
	print('creating {0} MB sparse file'.format(4*ar.s))
	task_sz = ar.s * 1024 * 1024
	
	with open('gen.out', 'wb') as f:
		f.seek(task_sz*4-3)
		f.write(b'eof')
		print('flushing')

	workers = []
	for v0, ofs in [
		[0x00000000, 0*task_sz],
		[0x40000000, 1*task_sz],
		[0x80000000, 2*task_sz],
		[0xc0000000, 3*task_sz]
		]:
		worker = make_worker(target=writer, args=(task_sz,v0,ofs,))
		workers.append(worker)
		worker.start()

	# https://bugs.python.org/issue1167930
	while True:
		time.sleep(0.1)
		if not is_busy():
			break


"""
# set up a ramdisk served over tcp
apk add nbd nbd-client
modprobe nbd
head -c $((64*1024*1024)) /dev/zero > /dev/shm/vdisk
mkfs.ext2 /dev/shm/vdisk
nbd-server 127.0.0.1:10809 /dev/shm/vdisk -C ''
nbd-client 127.0.0.1 10809 /dev/nbd0
mount -o sync /dev/nbd0 /media/slow/

# set speed limit
tc qdisc del dev lo root
tc qdisc add dev lo root handle 1: htb default 10
tc class add dev lo parent 1: classid 1:1 htb rate 1mbit burst 15k
tc filter add dev lo parent 1: protocol ip prio 1 u32 match ip sport 10809 0xffff flowid 1:1
tc filter add dev lo parent 1: protocol ip prio 1 u32 match ip dport 10809 0xffff flowid 1:1

# run this test
rm gen.out; ~/dev/sparsemulti.py thr
hexdump -C gen.out | grep '00 00 00 00 00 00 00 00' -B1
wc -c gen.out  # bytes apparent size
du -sk gen.out # kbytes disk consumption

# unmount and delete when done
umount /media/slow
killall nbd-client
qemu-nbd -d /dev/nbd0
rm /dev/shm/vdisk

# example output, a sparse file populated by multiple writers in parallel
00009ff0  00 00 27 fc 00 00 27 fd  00 00 27 fe 00 00 27 ff  |..'...'...'...'.|
0000a000  00 00 00 00 00 00 00 00  00 00 00 00 00 00 00 00  |................|
--
10009ff0  40 00 27 fc 40 00 27 fd  40 00 27 fe 40 00 27 ff  |@.'.@.'.@.'.@.'.|
1000a000  00 00 00 00 00 00 00 00  00 00 00 00 00 00 00 00  |................|
--
20009ff0  80 00 27 fc 80 00 27 fd  80 00 27 fe 80 00 27 ff  |..'...'...'...'.|
2000a000  00 00 00 00 00 00 00 00  00 00 00 00 00 00 00 00  |................|
--
30009ff0  c0 00 27 fc c0 00 27 fd  c0 00 27 fe c0 00 27 ff  |..'...'...'...'.|
3000a000  00 00 00 00 00 00 00 00  00 00 00 00 00 00 00 00  |................|
*
3ffffff0  00 00 00 00 00 00 00 00  00 00 00 00 00 65 6f 66  |.............eof|

1073741824 gen.out  # bytes apparent size
174	gen.out         # kbytes disk consumption
"""
